import base64
from nacl import utils as nacl_utils, secret, pwhash
import cenotes.exceptions
from cenotes.models import create_new_note, fetch_note
from cenotes.utils.other import enforce_bytes, safe_decryption


def generate_random_chars(size=32):
    return nacl_utils.random(size)


def generate_url_safe_pass(size=32):
    return base64.urlsafe_b64encode(generate_random_chars(size)).decode()


@enforce_bytes(kwargs_names="password")
def craft_key_from_password(password):
    kdf = pwhash.kdf_scryptsalsa208sha256
    salt = nacl_utils.random(pwhash.SCRYPT_SALTBYTES)
    ops = pwhash.SCRYPT_OPSLIMIT_SENSITIVE
    mem = pwhash.SCRYPT_MEMLIMIT_SENSITIVE
    return kdf(secret.SecretBox.KEY_SIZE, password, salt,
               opslimit=ops, memlimit=mem)


def craft_secret_box(key):
    return secret.SecretBox(key)


@enforce_bytes(kwargs_names="what")
def url_safe_encode(what):
    return base64.urlsafe_b64encode(what).decode()


@safe_decryption
@enforce_bytes(kwargs_names="what")
def url_safe_decode(what):
    return base64.urlsafe_b64decode(what)


@enforce_bytes(kwargs_names="what")
def encrypt_with_box(what, secret_box):
    return secret_box.encrypt(what)


@enforce_bytes(nof_args=2, kwargs_names=["what", "key"])
def encrypt_with_key(what, key):
    return encrypt_with_box(what, craft_secret_box(key))


@enforce_bytes(kwargs_names="what")
def encrypt_with_password(what, password):
    key = craft_key_from_password(password)
    return encrypt_with_key(what, key), key


@safe_decryption
@enforce_bytes(kwargs_names="what")
def decrypt_with_box(what, secret_box):
    return secret_box.decrypt(what)


@enforce_bytes(kwargs_names="what")
def decrypt_with_key(what, key):
    return decrypt_with_box(what, craft_secret_box(key))


def encrypt_note(note, password=None):
    if not note:
        raise cenotes.exceptions.InvalidUsage("Note payload cannot be empty")

    password = (password or "").encode() or generate_random_chars()

    ciphertext, key = encrypt_with_password(note, password)

    # needs to be url safe so we can share it around
    return ciphertext, url_safe_encode(key)


def craft_url_safe_encrypted_payload(cen_parameters, key=None):
    payload, encoded_key = encrypt_note(cen_parameters.plaintext,
                                        key or cen_parameters.key)
    return url_safe_encode(payload), encoded_key


def create_encrypted_note(cen_parameters, key=None):
    payload, encoded_key = encrypt_note(cen_parameters.plaintext,
                                        key or cen_parameters.key)
    return create_new_note(cen_parameters, payload), encoded_key


@safe_decryption
def decrypt_payload(payload, key):
    return decrypt_with_key(payload, url_safe_decode(key))


@safe_decryption
def decrypt_note(note_id, key):
    return decrypt_payload(fetch_note(int(note_id)), key)



import json
from cenotes.utils import CENParams


def craft_response(enote, error, key, payload, plaintext, success):
    return {"success": success if not error else False,
            "error": error,
            "key": key,
            "plaintext": plaintext,
            "payload": payload,
            "expiration_date": enote.iso_expiration_date if enote else "",
            "max_visits": enote.max_visits if enote else ""}


def craft_json_response(
        success=True, error="", enote=None, plaintext="", key="", payload=""):

    return json.dumps(
        craft_response(enote, error, key, payload, plaintext, success))


def get_request_params(request_params):
    return CENParams(**request_params)



import click
from cenotes import create_app


@click.command()
@click.option('--settings', show_default=True, type=click.Choice(
    ["Production", "Development", "Testing"]), default="Production")
@click.option("--host", type=str, default="localhost", show_default=True)
@click.option("--port", type=int, default=8080, show_default=True)
def main(settings, host, port):
    create_app(app_settings="cenotes.config_backend.{0}"
               .format(settings)).run(host=host, port=port)


if __name__ == '__main__':
    main()



from functools import wraps
from nacl.exceptions import CryptoError
from sqlalchemy.orm.exc import NoResultFound
from cenotes.exceptions import InvalidKeyORNoteError


def make_type(mtype, *args):
    t_args = []
    accepted_collections = (list, tuple, set)
    similar_type = tuple if mtype is list else list
    for x in args:
        if not isinstance(x, accepted_collections):
            t_args.extend([x])
        else:
            t_args.extend(similar_type(x))
    return mtype(t_args)


def make_tuple(*args):
    return make_type(tuple, *args)


def enforce_bytes(nof_args=1, kwargs_names=tuple(["test"])):
    def byte_force(what):
        try:
            return what.encode()
        except AttributeError:
            return what

    def enforcer(func):
        if set(make_tuple(kwargs_names)) - set(func.__code__.co_varnames):
            raise SyntaxWarning("Wrong kwarg names in decorator!")

        @wraps(func)
        def enforce(*args, **kwargs):
            new_args = [arg if i >= nof_args else byte_force(arg)
                        for i, arg in enumerate(args)]
            new_kwargs = {
                key: val if key not in kwargs_names else byte_force(val)
                for key, val in kwargs.items()
            }
            return func(*new_args, **new_kwargs)
        return enforce
    return enforcer


def safe_decryption(func):
    @wraps(func)
    def safe_decrypt(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (CryptoError, UnicodeDecodeError, NoResultFound,
                ValueError, TypeError) as err:
            raise InvalidKeyORNoteError(err)
    return safe_decrypt



import traceback
from flask import current_app


class CenotesError(Exception):
    base_text = "Cenotes Error"

    @staticmethod
    def should_traceback():
        return any(map(current_app.config.get,
                       ("DEBUG", "DEVELOPMENT", "TESTING")))

    def __str__(self, *args, **kwargs):
        if self.should_traceback():
            return "{0}: {1}".format(
                self.base_text,
                traceback.extract_tb(super(Exception, self).__traceback__))
        else:
            return self.base_text


class InvalidUsage(CenotesError):
    base_text = "Invalid usage"

    def __init__(self, *args, **kwargs):
        self.base_text = ",".join(args or kwargs.values())


class InvalidKeyORNoteError(CenotesError):
    base_text = "Invalid key or note not found"



from flask import Blueprint
from cenotes.exceptions import InvalidUsage, InvalidKeyORNoteError
from cenotes.utils.api import craft_json_response

error_bp = Blueprint('error_handlers', __name__)


@error_bp.app_errorhandler(404)
def not_found(error):
    return craft_json_response(error="Not found"), 404


@error_bp.app_errorhandler(400)
def bad_request(error):
    return craft_json_response(error="Bad request"), 404


@error_bp.app_errorhandler(InvalidUsage)
def invalid_usage(error):
    return craft_json_response(error=str(error)), 400


@error_bp.app_errorhandler(InvalidKeyORNoteError)
def invalid_key_or_note(error):
    return craft_json_response(error=str(error)), 400



from flask import Blueprint, request, current_app
from cenotes.utils import crypto as cu_crypto, api as capi
from cenotes.exceptions import InvalidKeyORNoteError

notes_bp = Blueprint('notes', __name__, url_prefix='/notes')


def _generate_decrypt_response(enote_id_or_payload, key):
    enote_id_or_payload = cu_crypto.url_safe_decode(enote_id_or_payload)
    try:
        enote_id = cu_crypto.decrypt_with_box(
            enote_id_or_payload, current_app.server_box)
    except InvalidKeyORNoteError:
        # if we can't decrypt it, we assume it's a payload instead of id
        enote_id = None
    if enote_id:
        plaintext = cu_crypto.decrypt_note(enote_id, key).decode()
    else:
        plaintext = cu_crypto.decrypt_payload(
            enote_id_or_payload, key).decode()
    return capi.craft_json_response(plaintext=plaintext), 200


@notes_bp.route("/", methods=["POST"])
def decrypt_json_note():
    request_params = request.get_json(silent=True) or {}
    cen_parameters = capi.get_request_params(request_params)
    enote_id_or_payload = cen_parameters.payload
    key = cen_parameters.key

    return _generate_decrypt_response(enote_id_or_payload, key)


@notes_bp.route("/<enote_id_or_payload>/<key>", methods=["GET"])
def decrypt_note(enote_id_or_payload, key):
    return _generate_decrypt_response(enote_id_or_payload, key)


@notes_bp.route("/<enote_id>/<key>", methods=["PATCH"])
@notes_bp.route("/<enote_id>", methods=["PATCH"])
def modify_note(enote_id, key):
    pass


@notes_bp.route("/<enote_id>/<key>", methods=["DELETE"])
@notes_bp.route("/<enote_id>", methods=["DELETE"])
def delete_note(enote_id, key):
    pass


@notes_bp.route("/encrypt/", methods=["POST"])
@notes_bp.route("/encrypt/<key>/", methods=["POST"])
def encrypt_note(key=None):
    request_params = request.get_json(silent=True) or {}
    cen_parameters = capi.get_request_params(request_params)

    if cen_parameters.no_store:
        new_note, final_key = cu_crypto.craft_url_safe_encrypted_payload(
            cen_parameters, key)
        return capi.craft_json_response(payload=new_note, key=final_key), 200

    else:
        new_note, final_key = cu_crypto.create_encrypted_note(
            cen_parameters, key)
        encrypted_id = cu_crypto.url_safe_encode(
            cu_crypto.encrypt_with_box(str(new_note.id),
                                       current_app.server_box))
        return capi.craft_json_response(
            payload=encrypted_id, enote=new_note, key=final_key), 200



class CENParams(object):
    def __init__(self, plaintext=None, key=None, expiration_date=None,
                 visits_count=None, max_visits=None, no_store=False,
                 payload=None, **kwargs):
        self.plaintext = plaintext
        self.key = key
        self.payload = payload
        self.expiration_date = expiration_date
        self.visits_count = visits_count
        self.max_visits = max_visits
        self.no_store = no_store
